import mnist
import scipy.misc
import numpy as np
import pandas as pd
from PIL import Image
import os

#Params of the model
main_smooth_cost_val=0.8
smoothness_cost_vals_for_roc=[10,1.2,0.8,0.7,0.6,0.2] #Make sure that you always put the main_smooth_cost_val inside this list
compatibility_cost=2

flip_chance=0.02

sample_images=10
test_images=10
num_images=sample_images+test_images

meanfield_epochs=10
noisefile='NoiseCoordinates.csv'
orderfile='UpdateOrderCoordinates.csv'
initialparamsfile='InitalParametersModel.csv'
energyfile='EnergyValues.csv'
energysampfile='EnergySamples.csv'
sampledenoisedfile='SampleDenoised.csv'
testdenoisedfile='TestDenoised.csv'
ROCfiletest='ROCTestData.csv'

#Energy function
def energycalc(pi_mat, inp_x,my_smoothness_cost,my_compatibility_cost):
    temp_mat=pi_mat * np.log(pi_mat+np.power(10.,-10.)) + (1 - pi_mat) * np.log(1 - pi_mat + np.power(10.,-10.))
    entropy = np.sum(temp_mat)
    expec_mat = 2 * pi_mat - 1

    log_p_first_term = 0
    for i in range(pi_mat.shape[0]):
        for j in range(pi_mat.shape[1]):

            if j + 1 < pi_mat.shape[1]:
                log_p_first_term = log_p_first_term + expec_mat[i, j] * expec_mat[i, j + 1]
            if i + 1 < pi_mat.shape[0]:
                log_p_first_term = log_p_first_term + expec_mat[i, j] * expec_mat[i + 1, j]
    log_p = log_p_first_term * my_smoothness_cost + my_compatibility_cost * np.sum(expec_mat * inp_x)

    return entropy - log_p

#Random Noise generator function
def random_noise_loc_generator(h=28,w=28,chance=0.02,num_image=500,outcsv='NoiseCoordinates.csv',col_prefix='Noise bit'):
    noisy_bits_num = int(chance * h * w)
    column_names=['Row Description'] + [ col_prefix+' '+str(x) for x in range(noisy_bits_num) ]
    df=pd.DataFrame(columns=column_names)

    np_row_noise=np.zeros((num_image,noisy_bits_num),dtype=int)
    np_col_noise = np.zeros((num_image, noisy_bits_num),dtype=int)
    for ctr in range(num_image):
        rand_locs = np.random.permutation(h * w)[:noisy_bits_num]
        my_rand_i = (rand_locs / h).astype(int)
        my_rand_j = (rand_locs - h * my_rand_i).astype(int)
        np_row_noise[ctr,:]=my_rand_i
        np_col_noise[ctr,:]=my_rand_j
        df.loc[2 * ctr] = ['Image ' + str(ctr) + ' Row'] + [str(x) for x in my_rand_i.tolist()]
        df.loc[2 * ctr + 1] = ['Image ' + str(ctr) + ' Column'] + [str(x) for x in my_rand_j.tolist()]
    df.to_csv(outcsv,index=False)
    return np_row_noise,np_col_noise

def random_noise_loc_loader(incsv='NoiseCoordinates.csv'):
    df=pd.read_csv(incsv,index_col=False)
    df.drop('Row Description',axis=1,inplace=True)
    loc_mat=df.values
    images_num=int(loc_mat.shape[0]/2)
    np_row_noise=loc_mat[[2*x for x in range(images_num)],:]
    np_col_noise=loc_mat[[2*x+1 for x in range(images_num)],:]
    return np_row_noise, np_col_noise


#Loading the MNist Data
#Source: https://github.com/datapythonista/mnist
images = mnist.train_images()
images=images.astype(float)/255
binarize_im = lambda im: (im>=0.5).astype(float)-(im<0.5).astype(float)#-1,1 bits!
binarized_ims= binarize_im(images)
height=binarized_ims.shape[1]
width=binarized_ims.shape[2]


if os.path.exists(noisefile):
    (np_row_noise, np_col_noise) = random_noise_loc_loader(incsv=noisefile)
else:
    (np_row_noise, np_col_noise) = random_noise_loc_generator(h=height,w=width,chance=flip_chance,
                                                              num_image=num_images,outcsv=noisefile)

if os.path.exists(orderfile):
    (np_row_order, np_col_order) = random_noise_loc_loader(incsv=orderfile)
else:
    (np_row_order, np_col_order) = random_noise_loc_generator(h=height,w=width,chance=1,col_prefix='Pixel',
                                                              num_image=num_images,outcsv=orderfile)

FPR_vals=[]
TPR_vals=[]
for smooth_ctr,smoothness_cost in enumerate(smoothness_cost_vals_for_roc):

    energy_mat=np.zeros((num_images,meanfield_epochs+1))
    original_binary_images=binarized_ims[:num_images,:,:]
    noisy_binary_images=np.zeros(original_binary_images.shape)
    reconstructed_images=np.zeros(original_binary_images.shape)
    accuracy_vals=np.zeros(num_images)
    for curr_im_idx in range(num_images):
        image=binarized_ims[curr_im_idx,:,:]

        curr_noise_rows=np_row_noise[curr_im_idx,:].astype(int).tolist()
        curr_noise_cols=np_col_noise[curr_im_idx,:].astype(int).tolist()
        flip_locs=np.zeros((height,width))
        flip_locs[curr_noise_rows,curr_noise_cols]=1
        flip_locs=flip_locs.astype(bool)

        X=image*(1-2*flip_locs.astype(float))# 1-2*0=1, 1-2*1=-1
        noisy_binary_images[curr_im_idx, :, :] = X

        if os.path.exists(initialparamsfile):
            pi_matrix = np.loadtxt(initialparamsfile,delimiter=',')
        else:
            pi_matrix = np.random.rand(height, width)
            np.savetxt(initialparamsfile, pi_matrix, delimiter=',',fmt='%s')


        energy_mat[curr_im_idx, 0] = energycalc(pi_matrix, X,smoothness_cost,compatibility_cost)
        #print('Initial Energy \t: ', energy_mat[curr_im_idx, 0] )

        for myiter in range(meanfield_epochs):
            rand_i=np_row_order[curr_im_idx,:]
            rand_j=np_col_order[curr_im_idx,:]

            for garbage,myctr in enumerate(range(height*width)):
                i=rand_i[myctr]
                j=rand_j[myctr]
                neighbors=[(i+eps[0],j+eps[1]) for eps in [(1,0),(-1,0),(0,1),(0,-1)] if
                           (i+eps[0]>=0 and i+eps[0]<height and j+eps[1]>=0 and j+eps[1]<width)]
                neighbors_i=[y[0] for y in neighbors]
                neighbors_j=[y[1] for y in neighbors]
                pi_neighbors=pi_matrix[neighbors_i,neighbors_j]
                exp_twice_sum_var=np.exp(2*  ( smoothness_cost*np.sum(2*pi_neighbors-1) + compatibility_cost*X[i,j] ))
                pi_matrix[i,j]=exp_twice_sum_var/(exp_twice_sum_var+1)

            energy_mat[curr_im_idx, myiter+1] = energycalc(pi_matrix, X,smoothness_cost,compatibility_cost)
            #print('Energy at step ', myiter, '\t: ', energy_mat[curr_im_idx, myiter+1])


        print('Image ', curr_im_idx,' finished.')

        #print(pi_matrix)
        map_im=binarize_im(pi_matrix)
        reconstructed_images[curr_im_idx,:,:]=map_im
        accuracy_vals[curr_im_idx]=np.sum((map_im==image).astype(float))/(height*width)

        #print('MAP Accuracy: ', np.sum((map_im==image).astype(float))/(height*width))
        #print('X and image incompatibility: ', 1-np.sum((image==X).astype(float))/(height*width))
        #print('Posterior Change: ', 1-np.sum((map_im==X).astype(float))/(height*width))
    if smoothness_cost==main_smooth_cost_val:
        np.savetxt(energyfile, energy_mat[sample_images:(sample_images+2),:2], delimiter=',',fmt='%s')
        np.savetxt(energysampfile, energy_mat[:sample_images,:], delimiter=',', fmt='%s')

    black_white_im=lambda inp: (255*(1+inp)/2).astype(int)
    original_binary_images=black_white_im(original_binary_images)
    noisy_binary_images=black_white_im(noisy_binary_images)
    reconstructed_images=black_white_im(reconstructed_images)

    orig_noisy_reconst_im=lambda idx: np.concatenate((original_binary_images[idx,:,:],noisy_binary_images[idx,:,:],
                                                      reconstructed_images[idx,:,:]),axis=0)

    if smoothness_cost == main_smooth_cost_val:
        np.savetxt(sampledenoisedfile, np.concatenate([(reconstructed_images[myidx,:,:]/255).astype(int) for myidx in range(sample_images)],axis=1),
                   delimiter=',', fmt='%s')
        np.savetxt(testdenoisedfile,np.concatenate([(reconstructed_images[myidx,:,:]/255).astype(int) for myidx in range(sample_images,num_images)],axis=1),
                   delimiter=',', fmt='%s')

        #sample figure
        samp_figure=np.concatenate([orig_noisy_reconst_im(myidx) for myidx in range(sample_images)],axis=1)
        im = scipy.misc.toimage(samp_figure, mode="L")
        im.save("SampleReconstruction.png", mode="L")

        #Test figure
        samp_figure = np.concatenate([np.concatenate((original_binary_images[idx,:,:],noisy_binary_images[idx,:,:]),axis=0)
                                      for idx in range(sample_images,num_images)], axis=1)
        im = scipy.misc.toimage(samp_figure, mode="L")
        im.save("TestOriginalAndNoisy.png", mode="L")

    #The ROC curve
    ref_vals = 1-(original_binary_images==noisy_binary_images).astype(float)
    predicted_vals= 1-(reconstructed_images==noisy_binary_images).astype(float)

    #ROC just for the test images
    ref_vals=ref_vals[sample_images:,:,:]
    predicted_vals=predicted_vals[sample_images:,:,:]

    FP= np.sum(predicted_vals * (1-ref_vals))
    FN= np.sum((1-predicted_vals)*ref_vals)
    TP= np.sum(predicted_vals * ref_vals)
    TN= np.sum((1-predicted_vals) * (1-ref_vals))

    FPR=FP/(FP+TN)
    TPR=TP/(TP+FN)

    FPR_vals.append(FPR)
    TPR_vals.append(TPR)

ROCTable=np.zeros((len(FPR_vals),2))
ROCTable[:,0]=np.array(FPR_vals)
ROCTable[:,1]=np.array(TPR_vals)
np.savetxt(ROCfiletest, ROCTable, delimiter=',',fmt='%s')

